import typing

from pyiron_workflow.nodes.function import Function
from pyiron_workflow.nodes.macro import Macro
from pyiron_workflow.nodes.transform import DataclassNode
from pyiron_workflow.node import Node


def get_color(node: Node, theme: typing.Literal['light']):

    if theme == 'light':
        return light_mode(node)
    else:
        raise ValueError(f'Theme must be one of ("light",) but got {theme}')

def light_mode(node: Node):
    if isinstance(node, Function):
        color_light_green = "#a2ea9f"
        return color_light_green
    elif isinstance(node, Macro):
        color_light_orange = "#eacf9f"
        return color_light_orange
    elif isinstance(node, DataclassNode):
        color_light_purple = "#cb9fea"
        return color_light_purple
    else:
        return node.color



from pyiron_workflow.type_hinting import type_hint_to_tuple, valid_value
from pyiron_workflow.channels import NotData
from pyiron_workflow.node import Node
from pyironflow.themes import get_color
import importlib
import typing
import warnings
import types
import math

NODE_WIDTH = 240

def get_import_path(obj):
    module = obj.__module__ if hasattr(obj, "__module__") else obj.__class__.__module__
    # name = obj.__name__ if hasattr(obj, "__name__") else obj.__class__.__name__
    name = obj.__name__ if "__name__" in dir(obj) else obj.__class__.__name__
    qualname = obj.__qualname__ if "__qualname__" in dir(obj) else obj.__class__.__qualname__

    warnings.simplefilter('error', UserWarning)
    if qualname != name:
        warnings.warn("Node __name__ does not match __qualname__ which may lead to unexpected behavior. To avoid this, ensure the node is NOT nested inside subclasses within the module.")

    path = f"{module}.{name}"
    if path == "numpy.ndarray":
        path = "numpy.array"
    return path

def dict_to_node(dict_node: dict, live_children: dict = None, reload=False) -> Node:
    """Convert dict spec of node back to Node object."""
    if live_children is None:
        live_children = {}
    data = dict_node['data']
    label = dict_node['id']
    node_id = data['python_object_id']
    # Check whether a node of the same label already exists in the underlying
    # workflow and whether it is the same object (by python id).  If so, return
    # that instance back so that the widget can avoid double adding the same
    # node and node data caches still work.
    if id(node := live_children.get(label, None)) != node_id:
        node = get_node_from_path(data['import_path'], reload=reload)(label=label)
    # if updating the workflow disconnect all edges here and let dict_to_edge
    # rebuild them so as to not keep edges in the underlying workflow that have
    # been removed in the GUI.
    node.inputs.disconnect()
    node.outputs.disconnect()
    if 'position' in dict_node:
        x, y = dict_node['position'].values()
        node.position = (x, y)
        # print('position exists: ', node.label, node.position)
    else:
        print('no position: ', node.label)
    if 'target_values' in data:
        target_values = data['target_values']
        target_labels = data['target_labels']
        for k, v in zip(target_labels, target_values):
            if v not in ('NonPrimitive', 'NotData', ''):
                type_hint = node.inputs[k].type_hint
                # JS gui can return input values like 2.0 as int, breaking type hints
                # so check here if the type hint is a float, but convert only if losslessly possible
                if isinstance(v, int) and not valid_value(v, type_hint) \
                        and valid_value(float(v), type_hint) and v == float(v):
                    v = float(v)
                node.inputs[k].value = v

    return node

def dict_to_edge(dict_edge, nodes):
    out = nodes[dict_edge['source']].outputs[dict_edge['sourceHandle']]
    inp = nodes[dict_edge['target']].inputs[dict_edge['targetHandle']]
    inp.connect(out)

    return True

def is_primitive(obj):
    primitives = (bool, str, int, float, type(None))
    return isinstance(obj, primitives)

def get_node_values(channel_dict):
    values = []
    for k, v in channel_dict.items():
        value = v.value
        if isinstance(value, NotData):
            value = 'NotData'
        elif not is_primitive(value):
            value = 'NonPrimitive'

        # JSON does not understand nan or infinity and JSON.parse will crash
        # the react front end on encountering it
        if isinstance(value, float) and not math.isfinite(value):
            value = None
        values.append(value)

    return values

def _get_generic_type(t):
    non_none_types = [arg for arg in t.__args__ if arg is not type(None)]
    return float if float in non_none_types else non_none_types[0]


def _get_type_name(t):
    primitive_types = (bool, str, int, float, type(None))
    if t is None:
        return 'None'
    elif t in primitive_types:
        return t.__name__
    else:
        return 'NonPrimitive'


def get_node_types(node_io):
    node_io_types = list()
    for k in node_io.channel_dict:
        type_hint = node_io[k].type_hint
        if isinstance(type_hint, (types.UnionType, typing._UnionGenericAlias)):
            type_hint = _get_generic_type(type_hint)

        node_io_types.append(_get_type_name(type_hint))
    return node_io_types


def get_node_position(node):
    if 'position' in dir(node):
        x, y = node.position
    else:
        x, y = 0, 0
    return {'x': x, 'y': y}


def get_node_dict(node, key=None):
    n_inputs = len(list(node.inputs.channel_dict.keys()))
    n_outputs = len(list(node.outputs.channel_dict.keys()))
    if n_outputs > n_inputs:
        node_height = 30 + (16*n_outputs) + 10
    else:
        node_height = 30 + (16*n_inputs) + 10
    label = node.label
    if (node.label != key) and (key is not None):
        label = f'{node.label}: {key}'
    return {
        'id': node.label,
        'data': {
            'label': label,
            'source_labels': list(node.outputs.channel_dict.keys()),
            'target_labels': list(node.inputs.channel_dict.keys()),
            'import_path': get_import_path(node),
            'target_values': get_node_values(node.inputs.channel_dict),
            'target_types': get_node_types(node.inputs),
            'source_values': get_node_values(node.outputs.channel_dict),
            'source_types': get_node_types(node.outputs),
            'failed': str(node.failed),
            'running': str(node.running),
            'ready': str(node.outputs.ready),
            'python_object_id': id(node),
        },
        'position': get_node_position(node),
        'type': 'customNode',
        'style': {'padding': 5,
                  'background': get_color(node=node, theme='light'),
                  'borderRadius': '10px',
                  'width': f'{NODE_WIDTH}PX',
                  'width_unitless': NODE_WIDTH,
                  'height': f'{node_height}px',
                  'height_unitless': node_height},
        'targetPosition': 'left',
        'sourcePosition': 'right'
    }


def get_nodes(wf):
    nodes = []
    for k, v in wf.children.items():
        nodes.append(get_node_dict(v, key=k))
    return nodes


def get_node_from_path(import_path, log=None, reload=False):
    """Import a node from a file path.

    Be careful with `reload` as it will break type hints from pyiron_workflow.

    Args:
        import_path (str): where to import from
        log (???): where to log to
        reload (bool): whether to reload modules in case their source changed.

    Returns:
        node
    """
    # Split the path into module and object part
    module_path, _, name = import_path.rpartition(".")
    # Import the module
    try:
        module = importlib.import_module(module_path)
    except ModuleNotFoundError as e:
        log.append_stderr(e)
        return None

    if reload:
        # Reload the module
        try:
            importlib.reload(module)
        except ImportError as e:
            if log:
                log.append_stderr(e)
            return None

    # Get the object
    object_from_path = getattr(module, name)
    return object_from_path


def get_edges(wf):
    edges = []
    for ic, (out, inp) in enumerate(wf.graph_as_dict["edges"]["data"].keys()):
        out_node, out_port = out.split('/')[2].split('.')
        inp_node, inp_port = inp.split('/')[2].split('.')

        edge_dict = dict()
        edge_dict["source"] = out_node
        edge_dict["sourceHandle"] = out_port
        edge_dict["target"] = inp_node
        edge_dict["targetHandle"] = inp_port
        edge_dict["id"] = ic

        edges.append(edge_dict)
    return edges

def get_input_types_from_hint(node_input: dict):

    new_type = ""

    for listed_type in list(type_hint_to_tuple(node_input.type_hint)):
        if listed_type == None:
            listed_type = type(None)
        if listed_type.__name__ != "NoneType":
            new_type = new_type + listed_type.__name__ + "|"

    new_type = new_type[:-1]

    for listed_type in list(type_hint_to_tuple(node_input.type_hint)):
        if listed_type == None:
            listed_type = type(None)
        if listed_type.__name__ == "NoneType":
            if new_type != "":
                new_type = ": Optional[" + new_type + "]"

    return new_type

def create_macro(wf = dict, name = str, root_path='../pyiron_nodes/pyiron_nodes'):

    imports = list("")
    var_def = ""

    file = open(root_path + '/' + name + '.py', 'w')

    for i, (k, v) in enumerate(wf.children.items()):
        rest, n = get_import_path(v).rsplit('.', 1)
        new_import = "    from " + rest + " import " + n
        imports.append(new_import)
        list_inputs = list(v.inputs.channel_dict.keys())

        for j in list(v.inputs):
            if ((v.label + "__" + j.label) in list(wf.inputs.channel_dict.keys())):
                if str(j) == ("NOT_DATA" or "None"):
                    value = "None"
                elif type(j.value) == str:
                    value = "'" + j.value + "'"
                else:
                    value = str(j.value)
                var_def = var_def + v.label + "_" + j.label + get_input_types_from_hint(j)+ " = " + value + ", "

    var_def = var_def[:-2]    

    count = 0
    new_list = list("")
    for ic, (out, inp) in enumerate(wf.graph_as_dict["edges"]["data"].keys()):
        out_node, out_port = out.split('/')[2].split('.')
        inp_node, inp_port = inp.split('/')[2].split('.')
        new_list.append([out_node, inp_node, inp_port])


    file.write(
'''from pyiron_workflow import as_function_node, as_macro_node
from typing import Optional

@as_macro_node()
def ''' + name + '''(self, ''' + var_def + '''):
''')
    for j in imports:
        file.write(j + "\n")

    for i, (k, v) in enumerate(wf.children.items()):
        rest, n = get_import_path(v).rsplit('.', 1)
        file.write("    self." + v.label + " = " +  n + "()\n") 
    
    for i, (k, v) in enumerate(wf.children.items()):
        rest, n = get_import_path(v).rsplit('.', 1)
    
        node_def =""
    
        for j in list(wf.inputs.channel_dict.keys()):
            node_label, input_label =j.rsplit('__', 1)
            if v.label == node_label: 
                node_def = node_def + input_label + " = " + node_label + "_" + input_label+ ", "
        
        for p in new_list:
            if v.label == p[1]:
                node_def = node_def + p[2] + " = self."+ p[0] + ", "
        node_def = node_def[:-2]
        file.write("    self." + v.label + ".set_input_values" + "(" + node_def + ")\n") 
    
    rest_list = []
    for items in list(wf.outputs.channel_dict.keys()):
        rest, n = items.rsplit('__', 1)
        rest_list.append(rest)

    out_str = "    return "
    for strs in rest_list:
        out_str = out_str + "self." + strs + ", "

    file.write(out_str)
    print("\nSuccessfully created macro: " + root_path + '/' + name + '.py')
    file.close()

    return



from contextlib import contextmanager
import pathlib
from typing import Literal
from dataclasses import dataclass
from enum import Enum
import json
import sys
import inspect
import re

import anywidget
import traitlets
from IPython.core import ultratb
from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.formatters import TerminalFormatter

from pyiron_workflow import Workflow
from pyiron_workflow.node import Node
from pyiron_workflow.nodes.transform import DataclassNode
from pyiron_workflow.nodes.function import Function as FunctionNode
from pyiron_workflow.nodes.macro import Macro as MacroNode
from pyironflow.wf_extensions import (
    get_nodes,
    get_edges,
    get_node_from_path,
    dict_to_node,
    dict_to_edge,
    create_macro,
    NODE_WIDTH
)
from pyiron_workflow.channels import ChannelConnectionError
from pyiron_workflow.mixin.run import ReadinessError

__author__ = "Joerg Neugebauer"
__copyright__ = (
    "Copyright 2024, Max-Planck-Institut for Sustainable Materials GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.2"
__maintainer__ = ""
__email__ = ""
__status__ = "development"
__date__ = "Aug 1, 2024"

_CHANNEL_CONNECTION_REGEX = '.*/[^/]+/(.*)\.type_hint = (.*); /[^/]+/(.*)\.type_hint = (.*)$'
_CHANNEL_TYPE_REGEX = \
    "^The channel /[^/]/([^\w]+) cannot take the value .* not compliant with the type hint (.*)$"


@contextmanager
def FormattedTB():
    sys_excepthook = sys.excepthook
    sys.excepthook = ultratb.FormattedTB(mode="Verbose", color_scheme="Neutral")
    yield
    sys.excepthook = sys_excepthook


def highlight_node_source(node: Node) -> str:
    """Extract and highlight source code of a node.

    Supported node types are function node, dataclass nodes and 'graph creator'.

    Args:
        node (pyiron_workflow.node.Node): node to extract source from

    Returns:
        highlighted source code.
    """
    try:
        match node:
            case FunctionNode():
                code = inspect.getsource(node.node_function)
            case MacroNode():
                code = inspect.getsource(node.graph_creator)
            case DataclassNode():
                code = inspect.getsource(node.dataclass)
            case _:
                return "Function to extract code not implemented!"
        return highlight(code, PythonLexer(), TerminalFormatter())
    except OSError as e:
        if e.args[0] == "could not find class definition":
            return "Could not locate source code."
        raise


class GlobalCommand(Enum):
    """Types of commands pertaining to the full workflow."""

    RUN = "run"
    SAVE = "save"
    LOAD = "load"
    DELETE = "delete"

    def handle(self, widget: 'PyironFlowWidget'):
        """Execute command on widget."""
        match self:
            case GlobalCommand.RUN:
                widget.select_output_widget()
                widget.out_widget.clear_output()
                widget.display_return_value(widget.wf.run)

            case GlobalCommand.SAVE:
                widget.select_output_widget()
                widget.wf.save()
                print(f"Successfully saved in {widget.wf.label}.")

            case GlobalCommand.LOAD:
                widget.select_output_widget()
                try:
                    widget.wf.load()
                    widget.update()
                    print(f"Successfully loaded from {widget.wf.label}.")
                except FileNotFoundError:
                    widget.update()
                    print(f"Save file {widget.wf.label} not found!")

            case GlobalCommand.DELETE:
                widget.select_output_widget()
                widget.wf.delete_storage()
                print(f"Deleted {widget.wf.label}.")

@dataclass
class NodeCommand:
    """Specifies a command to run a node or selection of them."""

    command: Literal["source", "run", "delete_node", "macro", "reset"]
    node: str


def parse_command(com: str) -> GlobalCommand | NodeCommand:
    """Parses commands from GUI into the correct command class."""
    print("command: ", com)
    if "executed at" in com:
        return GlobalCommand(com.split(" ")[0])

    command_name, node_name = com.split(":")
    node_name = node_name.split("-")[0].strip()
    return NodeCommand(command_name, node_name)


class ReactFlowWidget(anywidget.AnyWidget):
    path = pathlib.Path(__file__).parent / "static"
    _esm = path / "widget.js"
    _css = path / "widget.css"
    nodes = traitlets.Unicode("[]").tag(sync=True)
    edges = traitlets.Unicode("[]").tag(sync=True)
    selected_nodes = traitlets.Unicode("[]").tag(sync=True)
    selected_edges = traitlets.Unicode("[]").tag(sync=True)
    commands = traitlets.Unicode("[]").tag(sync=True)
    # position and size of the current view on the graph in JS space
    view = traitlets.Unicode("{}").tag(sync=True)


@contextmanager
def GentleError(out, log):
    """Catch various exception from workflows and try to print nicer messages.

    Args:
        out: widget for "normal" output immediately visible to user
        log: widget for "logging" output only visible after a click
    """
    try:
        try:
            yield
        except ReadinessError as err:
            with out:
                print("The following node require inputs before you can run the graph:")
                def clean(s):
                    if s.startswith("inputs."):
                        s = s[len("inputs."):]
                    return s.replace("__", ".")
                unready_channels = [clean(k) for k, v in err.readiness_dict.items()
                                        if not v and k not in ("ready", "running", "failed")]
                print(*unready_channels, sep='\n')
            with log:
                sys.excepthook(*sys.exc_info())
        except ChannelConnectionError as err:
            with out:
                groups = re.match(_CHANNEL_CONNECTION_REGEX, err.args[0]).groups()
                if groups is not None and len(groups) == 4:
                    leftchannel, lefttype, rightchannel, righttype = groups
                    print(f"Error: Cannot connect {leftchannel} to {rightchannel}!\n"
                            f"Their types do not match {lefttype} != {righttype}.")
                else:
                    print("Error: Could not connect some edges because of type mismatch!")
            with log:
                sys.excepthook(*sys.exc_info())
        except TypeError as err:
            with out:
                groups = re.match(_CHANNEL_TYPE_REGEX, err.args[0])
                if groups is not None and len(groups) == 2:
                    channel, typehint = groups
                    print(f"Channel {channel} connected to wrong type! Should be {typehint}.")
            with log:
                sys.excepthook(*sys.exc_info())
    except Exception as e:
        print("Error:", e)
        with log:
            sys.excepthook(*sys.exc_info())
    finally:
        pass



class PyironFlowWidget:
    def __init__(
        self,
        root_path="../pyiron_nodes/pyiron_nodes",
        wf: Workflow = Workflow(label="workflow"),
        log=None,
        out_widget=None,
        reload_node_library=False,
    ):
        self.log = log
        self.out_widget = out_widget
        self.accordion_widget = None
        self.tree_widget = None
        self.gui = ReactFlowWidget(layout={'height': '100%'})
        self.wf = wf
        self.root_path = root_path
        self.reload_node_library = reload_node_library

        self.gui.observe(self.on_value_change, names="commands")

        self.update()

    def select_output_widget(self):
        """Makes sure output widget is visible if accordion is set."""
        if self.accordion_widget is not None:
            self.accordion_widget.selected_index = 1

    def display_return_value(self, func):
        from IPython.display import display
        with FormattedTB(), GentleError(self.out_widget, self.log):
            display(func())

    def on_value_change(self, change):

        self.out_widget.clear_output()

        error_message = ""

        with FormattedTB(), GentleError(self.out_widget, self.log):
            try:
                self.wf = self.get_workflow()
            except Exception as error:
                error_message = error
                raise

        if "done" in change["new"]:
            return

        import warnings

        with self.out_widget, warnings.catch_warnings(action="ignore"):
            match parse_command(change["new"]):
                case GlobalCommand() as command:
                    command.handle(self)
                case NodeCommand("macro", node_name):
                    self.select_output_widget()
                    create_macro(
                        self.get_selected_workflow(), node_name, self.root_path
                    )
                    if self.tree_widget is not None:
                        self.tree_widget.update_tree()

                case NodeCommand(command, node_name):
                    if node_name not in self.wf.children:
                        return
                    node = self.wf.children[node_name]
                    self.select_output_widget()
                    match command:
                        case "reset":
                            node.failed = False
                            node.running = False
                            if node.use_cache:
                                node._cached_inputs = {}
                            self.wf.failed = False
                            self.update_status()
                        case "source":
                            print(highlight_node_source(node))
                        case "run":
                            if error_message:
                                print(f"Could not run node {node_name}!")
                            else:
                                self.display_return_value(node.pull)
                        case "delete_node":
                            self.wf.remove_child(node_name)
                        case command:
                            print(f"ERROR: unknown command: {command}!")
                case unknown:
                    print(f"Command not yet implemented: {unknown}")

    def update(self):
        nodes = get_nodes(self.wf)
        edges = get_edges(self.wf)
        self.gui.nodes = json.dumps(nodes)
        self.gui.edges = json.dumps(edges)

    def update_status(self):
        temp_nodes = get_nodes(self.wf)
        temp_edges = get_edges(self.wf)
        self.wf = self.get_workflow()
        actual_nodes = get_nodes(self.wf)
        actual_edges = get_edges(self.wf)
        for i in range(len(actual_nodes)):
            actual_nodes[i]["data"]["failed"] = temp_nodes[i]["data"]["failed"]
            actual_nodes[i]["data"]["running"] = temp_nodes[i]["data"]["running"]
            actual_nodes[i]["data"]["ready"] = temp_nodes[i]["data"]["ready"]
        self.gui.nodes = json.dumps(actual_nodes)
        self.gui.edges = json.dumps(actual_edges)

    @property
    def react_flow_widget(self):
        return self.gui

    def place_new_node(self):
        """Find a suitable location in UI space for the newly added node.

        Exact layouting not required as this can be done in UI, but newly added
        nodes should be visible to the user and not completely overlap.

        FIXME: Probably this is better handled completely in UI by elk.
        """
        view = json.loads(self.gui.view)
        if view == {}:
            position = [0, 0]
        else:
            position = [
                    -view['x'] + 0.1 * view['height'],
                    -view['y'] + 0.9 * view['height'],
            ]

        def blocked():
            for node in self.wf.children.values():
                if 'position' in dir(node):
                    if node.position == tuple(position):
                        return True
            return False
        while blocked():
            position[0] += NODE_WIDTH + 10

        return tuple(position)

    def add_node(self, node_path, label):
        self.wf = self.get_workflow()
        node = get_node_from_path(node_path, log=self.log)
        node.position = self.place_new_node()
        if node is not None:
            self.log.append_stdout(f"add_node (reactflow): {node}, {label} \n")
            if label in self.wf.child_labels:
                self.wf.strict_naming = False

            self.wf.add_child(node(label=label))

            self.update()

    def get_workflow(self):
        wf = self.wf
        dict_nodes = json.loads(self.gui.nodes)
        for dict_node in dict_nodes:
            node = dict_to_node(dict_node, wf.children, reload=self.reload_node_library)
            if node not in wf.children.values():
                # new node appeared in GUI with the same name, but different
                # id, i.e. user removed and added something in place
                if node.label in wf.children:
                    # FIXME look at replace_child
                    wf.remove_child(node.label)
                wf.add_child(node)

        dict_edges = json.loads(self.gui.edges)
        for dict_edge in dict_edges:
            dict_to_edge(dict_edge, wf.children)

        return wf

    def get_selected_workflow(self):
        wf = Workflow("temp_workflow")
        dict_nodes = json.loads(self.gui.selected_nodes)
        node_labels = []
        for dict_node in dict_nodes:
            node = dict_to_node(dict_node)
            wf.add_child(node)
            node_labels.append(dict_node["data"]["label"])
            # wf.add_child(node(label=node.label))
        print("\nSelected nodes:")
        print(node_labels)

        nodes = wf.children
        dict_edges = json.loads(self.gui.selected_edges)
        subset_dict_edges = []
        edge_labels = []
        for edge in dict_edges:
            if edge["source"] in node_labels and edge["target"] in node_labels:
                subset_dict_edges.append(edge)
                edge_labels.append(edge["id"])
        print("\nSelected edges:")
        print(edge_labels)

        for dict_edge in subset_dict_edges:
            dict_to_edge(dict_edge, nodes)

        return wf



# Internal init
from ._version import get_versions

# Set version of pyiron_base
__version__ = get_versions()["version"]

try:
    from .pyironflow import PyironFlow
except FileNotFoundError:
    print("WARNING: could not import PyironFlow, likely because js sources "
          "are not build and we are in build env that just tries to get the "
          "version number.")



from ipytree import Tree, Node
from ipywidgets import HBox, VBox, Button
from pathlib import Path
import ast

from dataclasses import dataclass

__author__ = "Joerg Neugebauer"
__copyright__ = (
    "Copyright 2024, Max-Planck-Institut for Sustainable Materials GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.2"
__maintainer__ = ""
__email__ = ""
__status__ = "development"
__date__ = "Aug 1, 2024"


# Note: available icons and types in ipytree
# - style_values = ["warning", "danger", "success", "info", "default"]
# - icons: https://fontawesome.com/v5/search?q=node&o=r (version 5) appears to work

WELL_KNOWN_NODE_WRAPPERS = (
    "as_function_node",
    "as_macro_node",
    "as_dataclass_node",
    "Workflow.wrap.as_function_node",
    "Workflow.wrap.as_macro_node",
    "Workflow.wrap.as_dataclass_node",
)


@dataclass
class FunctionNode:
    name: str
    path: str | Path


@dataclass
class DataClassNode:
    name: str
    path: str | Path


def get_rel_path_for_last_occurrence(path: Path, relpath_start: str) -> int:
    if relpath_start in path.parts:
        # Reverse the list and find the first (last in original list) occurrence
        reversed_parts = path.parts[::-1]  # this does not modify the original list
        last_occurrence = len(path.parts) - 1 - reversed_parts.index(relpath_start)

        rel_path = Path(*path.parts[last_occurrence:])
        rel_path_no_ext = rel_path.with_suffix("")
        return rel_path_no_ext


class TreeView:
    def __init__(
        self, root_path="../pyiron_nodes/pyiron_nodes", flow_widget=None, log=None
    ):
        """
        This function generates and returns a tree view of nodes starting from the
        root_path directory.

        Params:
        ------
        root_path : str or Path, optional
            The root directory path from which the tree starts.
            Defaults to '../pyiron_nodes/pyiron_nodes'.

        Return:
        ------
        tree : Tree object
            A tree view object with nodes added to it.
        """
        import copy

        self.path = copy.copy(root_path)
        if isinstance(self.path, str):
            self.path = Path(root_path)

        self.flow_widget = flow_widget
        self.log = log  # logging widget

        self.refresh_button = Button(
            description="Refresh", disabled=False, button_style="info"
        )

        self.tree = Tree(stripes=True)
        self.add_nodes(self.tree, parent_node=self.path)

        self.refresh_button.on_click(self.update_tree)
        # the following flag is needed since handle click sends two signals, the first repeats the last one from the
        # previous click
        self._handle_click_is_last_event = True

        self.gui = VBox([self.refresh_button, self.tree])

    def update_tree(self, b=None):
        for tree_nodes in self.tree.nodes:
            self.tree.remove_node(tree_nodes)
        self.add_nodes(self.tree, parent_node=self.path)

    def handle_click(self, event):
        """
        This function handles click events by adding nodes to the selected object
        if it does not already have any nodes.

        Params:
        ------
        event : dict
            A dictionary representing the event object.

        Note:
        The event object should include the owner of the event (the object that was clicked),
        and the owner should have a 'nodes' property (a list of nodes) and a 'path' property (the path to the node).
        """
        if not self._handle_click_is_last_event:
            self._handle_click_is_last_event = True
            return None
        self._handle_click_is_last_event = False

        selected_node = event["owner"]
        # self.log.append_stdout(f'handle_click ({selected_node.path}, {selected_node.name}) \n')

        if selected_node.icon in ["codepen", "table"]:
            selected_node.on_click(selected_node)
        elif (len(selected_node.nodes)) == 0:
            self.add_nodes(selected_node, selected_node.path)

    def on_click(self, node):
        import os

        # self.log.append_stdout(f'on_click.add_node_init ({node.path}, {node.path.name}) \n')
        path = os.path.join(
            get_rel_path_for_last_occurrence(node.path.path, "pyiron_nodes"),
            node.path.name,
        )
        path_str = str(path).replace(os.sep, ".")
        if self.flow_widget is not None:
            # self.log.append_stdout(f'on_click.add_node ({str(path_str)}, {node.path.name}) \n')
            self.flow_widget.add_node(str(path_str), node.path.name)

    def add_nodes(self, tree, parent_node):
        """
        This function adds child nodes to a parent node in a tree. It assumes the input
        is an Abstract Syntax Tree (AST). It creates new nodes based on the attributes
        of the parent node, updates icon style based on the type of node and finally
        adds child nodes to the parent.

        Params:
        ------
        tree : ast
            The Abstract Syntax Tree

        parent_node : Node object
            The node of the AST to which child nodes must be added

        """

        for node in self.list_nodes(parent_node):
            name_lst = node.name.split(".")
            if len(name_lst) > 1:
                if "py" == name_lst[-1]:
                    node_tree = Node(name_lst[0])
                    node_tree.icon = "archive"  # 'file'
                    node_tree.icon_style = "success"
                else:
                    continue
            else:
                node_tree = Node(node.name)
                if isinstance(node, FunctionNode):
                    node_tree.icon = "codepen"  # 'file-code' # 'code'
                    node_tree.icon_style = "danger"
                elif isinstance(node, DataClassNode):
                    node_tree.icon = "table"  # 'file-code' # 'code'
                    node_tree.icon_style = "success"
                else:
                    node_tree.icon = "folder"  # 'info', 'copy', 'archive'
                    node_tree.icon_style = "warning"

            node_tree.path = node
            tree.add_node(node_tree)
            if self.on_click is not None:
                node_tree.on_click = self.on_click

            node_tree.observe(self.handle_click, "selected")

    def list_nodes(self, node: Path):
        """
        Return a list of child directories and python files of a given Path' node'.
        Child directories and python files starting with '.' or '_' are excluded.

        Parameters:
        node (Path): A directory or a python file.

        Returns:
        nodes (List[Path]): List of child directories and python files. For python file 'node',
          list_pyiron_nodes(node) is called and the paths are added.
        """
        node_path = node

        nodes = []
        if node.is_dir():
            for child in node_path.iterdir():
                if (
                    child.is_dir()
                    and not child.name.startswith(".")
                    and not child.name.startswith("_")
                ):
                    nodes.append(child)

            for child in node_path.glob("*.py"):
                if not child.name.startswith(".") and not child.name.startswith("_"):
                    nodes.append(child)

        elif node.is_file():
            for child in self.list_pyiron_nodes(node):
                nodes.append(child)

        return nodes

    @staticmethod
    def list_pyiron_nodes(file_name, decorators=WELL_KNOWN_NODE_WRAPPERS):
        """
        This function reads a Python code file and looks for any assignments
        to a list variable named 'nodes'. It then creates FunctionNode objects
        for each element in this list and returns all FunctionNodes in a list.

        Params:
        ------
        file_name : str
            Path to the python file to be analysed

        Returns:
        -------
        nodes : list of FunctionNode
            List of FunctionNodes extracted from the Python file
        """
        with open(file_name, "r") as file:
            tree = ast.parse(file.read())

        nodes = []

        def wrap_node(
            node: ast.ClassDef | ast.FunctionDef,
        ) -> FunctionNode | DataClassNode:
            match node:
                case ast.ClassDef():
                    node = DataClassNode(name=node.name, path=Path(file_name))
                case ast.FunctionDef():
                    node = FunctionNode(name=node.name, path=Path(file_name))
                case unknown:
                    assert False, (
                        f"wrap_node called with wrong ast node type: {unknown}!"
                    )
            nodes.append(node)

        def full_name(attr: ast.Attribute | ast.Name) -> str:
            """Build str rep of an arbitrarily nested attribute access."""
            if isinstance(attr, ast.Name):
                return attr.id
            parent = attr.value
            name = attr.attr
            return full_name(parent) + "." + name

        for node in ast.walk(tree):
            if isinstance(node, (ast.FunctionDef, ast.ClassDef)):
                for decorator in node.decorator_list:
                    # if decorator is called in node source, access the callable
                    if isinstance(decorator, ast.Call):
                        decorator = decorator.func
                    if not isinstance(decorator, (ast.Name, ast.Attribute)):
                        continue  # don't know how to handle decorator references that are not plain names or attributes
                    if full_name(decorator) in decorators:
                        wrap_node(node)
                        break

        return nodes



import ipywidgets as widgets
from pyironflow.treeview import TreeView
from pyironflow.reactflow import PyironFlowWidget
from IPython.display import display
from pyiron_workflow import Workflow

__author__ = "Joerg Neugebauer"
__copyright__ = (
    "Copyright 2024, Max-Planck-Institut for Sustainable Materials GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.2"
__maintainer__ = ""
__email__ = ""
__status__ = "development"
__date__ = "Aug 1, 2024"

class GUILayout:
    screen_width = None
    screen_height = None
    flow_widget_width = None
    flow_widget_height = None
    output_widget_width = None


class PyironFlow:
    def __init__(
            self,
            wf_list=None, root_path=None,
            gui_layout: GUILayout | None = None,
            flow_widget_ratio: float = 0.85,
            reload_node_library: bool = False,
    ):
        """

        Args:
            ...
            gui_layout (GUILayout): ignored
            flow_widget_ratio (float): fraction of the widget width that is reserved for the workflow view.
            reload_node_library (bool): allow the refresh button to reload node modules
        """
        if gui_layout is not None:
            warnings.warn("gui_layout is ignored for the widget size, use flow_widget_ratio to control "
                          "'output_widget_width'")
        # throw a warning; debate value limits
        flow_widget_ratio = max(min(flow_widget_ratio, 0.95), 0.05)

        # generate empty default workflow if workflow list is empty
        if wf_list is None:
            wf_list = []
        if len(wf_list) == 0:
            wf_list = [Workflow('workflow')]

        if root_path is None:
            try:
                import pyiron_nodes
                root_path = pyiron_nodes.__spec__.submodule_search_locations[0]
            except (ImportError, IndexError):
                root_path = ""

        self._flow_widget_factor = 1 / (1/flow_widget_ratio - 1)
        self.workflows = wf_list

        self.out_log = widgets.Output(layout={'border': '1px solid black', 'overflow': 'auto', })
        self.out_widget = widgets.Output(layout={'border': '1px solid black', 'overflow': 'auto', })
        self.wf_widgets = [PyironFlowWidget(
                wf=wf, root_path=root_path, log=self.out_log,
                out_widget=self.out_widget,
                reload_node_library=reload_node_library
            ) for wf in self.workflows]
        self.view_flows = self.view_flows()
        self.tree_view = TreeView(root_path=root_path, flow_widget=self.wf_widgets[0], log=self.out_log)
        self.accordion = widgets.Accordion(children=[self.tree_view.gui, self.out_widget, self.out_log],
                                           titles=['Node Library', 'Output', 'Logging Info'],
                                           layout={'border': '1px solid black',
                                                   'width': f'{int(100*(1-flow_widget_ratio))}%',
                                                   'flex': '1 0 auto',
                                                   'overflow': 'auto', })
        for widget in self.wf_widgets:
            widget.accordion_widget = self.accordion
            widget.tree_widget = self.tree_view

        self.gui = widgets.HBox([
            self.accordion,
            self.view_flows,
            # self.out_widget
        ],
            layout={
                'border': '1px solid black',
                'flex': '1 1 auto',
                'width': 'auto',
                'height': '75vh',
            })

    def get_workflow(self, tab_index=0):
        wf_widget = self.wf_widgets[tab_index]
        return wf_widget.get_workflow()

    def view_flows(self):
        tab = widgets.Tab(layout={'width': 'auto',
                                  'flex': f'{self._flow_widget_factor} 0 auto',
                                  'height': '100%'})
        tab.children = [self.display_workflow(index) for index, _ in enumerate(self.workflows)]
        tab.titles = [wf.label for wf in self.workflows]
        return tab

    def display_workflow(self, index: int):
        w = self.wf_widgets[index]
        return w.gui
